from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Sequence

from typing_extensions import Literal, Protocol, TypedDict

if TYPE_CHECKING:
    from ._descriptor import MimeBundleDescriptor


class UpdateData(TypedDict):
    method: Literal["update"]
    state: dict
    buffer_paths: list[list[int | str]]


class RequestStateData(TypedDict):
    method: Literal["request_state"]


class CustomData(TypedDict):
    method: Literal["custom"]
    content: dict  # Generic[ContentT] ... but only works with TypedDict in py311


class JupyterWidgetContent(TypedDict):
    comm_id: str
    data: UpdateData | RequestStateData | CustomData


class CommMessage(TypedDict):
    header: dict
    # typically UUID, must be unique per message
    msg_id: str
    msg_type: str
    parent_header: dict
    metadata: dict
    content: JupyterWidgetContent
    buffers: list[memoryview]


class MimeReprCallable(Protocol):
    """Protocol for _repr_mimebundle_.

    https://ipython.readthedocs.io/en/stable/config/integrating.html#more-powerful-methods

    > Should return a dictionary of multiple formats, keyed by mimetype, or a tuple of
    > two dictionaries: `data`, `metadata` (see Distribution metadata). If this returns
    > something, other `_repr_*_` methods are ignored. The method should take keyword
    > arguments `include` and `exclude`, though it is not required to respect them."
    """

    def __call__(
        self, include: Sequence[str], exclude: Sequence[str]
    ) -> dict | tuple[dict, dict]: ...


class AnywidgetProtocol(Protocol):
    """Anywidget classes have a MimeBundleDescriptor at `_repr_mimebundle_`."""

    _repr_mimebundle_: MimeBundleDescriptor


class WidgetBase(Protocol):
    """Widget subclasses with a custom message reducer."""

    def send(self, msg: str | dict | list, buffers: list[bytes]) -> None: ...

    def on_msg(
        self, callback: Callable[[Any, str | list | dict, list[bytes]], None]
    ) -> None: ...



from __future__ import annotations

import contextlib
import dataclasses
import pathlib
import typing

import psygnal

from ._descriptor import MimeBundleDescriptor

if typing.TYPE_CHECKING:  # pragma: no cover
    from ._protocols import WidgetBase

__all__ = ["dataclass", "widget", "MimeBundleDescriptor"]

_T = typing.TypeVar("_T")
T = typing.TypeVar("T")


def widget(
    *,
    esm: str | pathlib.Path,
    css: None | str | pathlib.Path = None,
    **kwargs: typing.Any,
) -> typing.Callable[[T], T]:
    """Decorator to register a widget class as a mimebundle.

    Parameters
    ----------
    esm : str | pathlib.Path
        The path or contents of an ES Module for the widget.
    css : None | str | pathlib.Path, optional
        The path or contents of a CSS file for the widget.
    **kwargs
        Additional keyword arguments to pass to the

    Returns
    -------
    Callable
        A decorator that registers the widget class as a mimebundle.
    """
    kwargs["_esm"] = esm
    if css is not None:
        kwargs["_css"] = css

    def _decorator(cls: _T) -> _T:
        setattr(cls, "_repr_mimebundle_", MimeBundleDescriptor(**kwargs))
        return cls

    return _decorator


# To preserve the signature of the decorated class.
# see: https://github.com/pyapp-kit/magicgui/blob/5e068f31eaeeb130f43c38727b25423cc3ea4de3/src/magicgui/schema/_guiclass.py#L145-L162
def __dataclass_transform__(
    *,
    eq_default: bool = True,
    order_default: bool = False,
    kw_only_default: bool = False,
    field_specifiers: tuple[type | typing.Callable[..., typing.Any], ...] = (()),
) -> typing.Callable[[_T], _T]:
    return lambda a: a


@__dataclass_transform__(field_specifiers=(dataclasses.Field, dataclasses.field))
def dataclass(
    cls: T | None = None,
    *,
    esm: str | pathlib.Path,
    css: None | str | pathlib.Path = None,
    **dataclass_kwargs: typing.Any,
) -> typing.Callable[[T], T]:
    """Turns class into a dataclass, makes it evented, and registers it as a widget.

    Parameters
    ----------
    cls : T | None
        The class to decorate.
    esm : str | pathlib.Path
        The path or contents of an ES Module for the widget.
    css : None | str | pathlib.Path, optional
        The path or contents of a CSS file for the widget.
    dataclass_kwargs : typing.Any
        Additional keyword arguments to pass to the dataclass decorator.

    Returns
    -------
    type
        The evented dataclass.

    Examples
    --------
    >>> @dataclass(esm="index.js")
    ... class Counter:
    ...     value: int = 0
    ...
    >>> counter = Counter()
    >>> counter.value = 1
    >>> counter
    """

    def _decorator(cls: T) -> T:
        cls = dataclasses.dataclass(cls, **dataclass_kwargs)  # type: ignore
        cls = psygnal.evented(cls)  # type: ignore
        cls = widget(esm=esm, css=css)(cls)
        return cls

    return _decorator(cls) if cls is not None else _decorator  # type: ignore


_ANYWIDGET_COMMAND = "_anywidget_command"

_AnyWidgetCommand = typing.Callable[
    [typing.Any, typing.Any, typing.List[bytes]],
    typing.Tuple[typing.Any, typing.List[bytes]],
]


def command(cmd: _AnyWidgetCommand) -> _AnyWidgetCommand:
    """Mark a function as a command for anywidget."""
    setattr(cmd, _ANYWIDGET_COMMAND, True)
    return cmd


_AnyWidgetCommandBound = typing.Callable[
    [typing.Any, typing.List[bytes]], typing.Tuple[typing.Any, typing.List[bytes]]
]


def _collect_commands(widget: WidgetBase) -> dict[str, _AnyWidgetCommandBound]:
    cmds: dict[str, _AnyWidgetCommandBound] = {}
    for attr_name in dir(widget):
        # suppressing silly assertion erro from ipywidgets _staticproperty
        # ref: https://github.com/jupyter-widgets/ipywidgets/blob/b78de43e12ff26e4aa16e6e4c6844a7c82a8ee1c/python/ipywidgets/ipywidgets/widgets/widget.py#L291-L297
        with contextlib.suppress(AssertionError):
            attr = getattr(widget, attr_name)
            if callable(attr) and getattr(attr, _ANYWIDGET_COMMAND, False):
                cmds[attr_name] = attr
    return cmds


def _register_anywidget_commands(
    widget: WidgetBase,
) -> None:
    """Register a custom message reducer for a widget if it implements the protocol."""
    # Only add the callback if the widget has any commands.
    cmds = _collect_commands(widget)
    if len(cmds) == 0:
        return None

    def handle_anywidget_command(
        self: WidgetBase, msg: str | list | dict, buffers: list[bytes]
    ) -> None:
        if not isinstance(msg, dict) or msg.get("kind") != "anywidget-command":
            return
        cmd = cmds[msg["name"]]
        response, buffers = cmd(msg["msg"], buffers)
        self.send(
            {
                "id": msg["id"],
                "kind": "anywidget-command-response",
                "response": response,
            },
            buffers,
        )

    widget.on_msg(handle_anywidget_command)



import pathlib
import sys
from unittest.mock import MagicMock, patch

import pytest
from anywidget._file_contents import FileContents
from anywidget._util import (
    get_repr_metadata,
    put_buffers,
    remove_buffers,
    try_file_contents,
)


def enable_hmr():
    return patch.dict("os.environ", {"ANYWIDGET_HMR": "1"}, clear=True)


def test_remove_and_put_buffers():
    mv1 = memoryview(b"test1")
    mv2 = memoryview(b"test2")
    state = {
        "plain": [0, "text"],  # should not get removed
        "x": {"ar": mv1},  # should result in an empty dict
        "y": {"shape": (10, 10), "data": mv1},
        "z": (mv1, mv2),  # tests tuple assignment
        "top": mv1,  # test a top level removal
        "deep": {"a": 1, "b": [0, {"deeper": mv2}]},
    }  # deeply nested
    plain = state["plain"]
    x = state["x"]
    y = state["y"]
    y_shape = y["shape"]
    state_before = state
    state, buffer_paths, buffers = remove_buffers(state)

    # check if buffers are removed
    assert "plain" in state
    assert "shape" in state["y"]
    assert "ar" not in state["x"]
    assert state["x"] == {}
    assert "data" not in state["y"]
    assert mv1 not in state["z"]
    assert mv1 not in state["z"]
    assert "top" not in state
    assert "deep" in state
    assert "b" in state["deep"]
    assert "deeper" not in state["deep"]["b"][1]

    # check that items that didn't need change aren't touched
    assert state is not state_before
    assert state["plain"] is plain
    assert state["x"] is not x
    assert state["y"] is not y
    assert state["y"]["shape"] is y_shape

    # check that the buffer paths really point to the right buffer
    for path, buffer in [
        (["x", "ar"], mv1),
        (["y", "data"], mv1),
        (["z", 0], mv1),
        (["z", 1], mv2),
        (["top"], mv1),
        (["deep", "b", 1, "deeper"], mv2),
    ]:
        assert path in buffer_paths, f"{path!r} not in path"
        index = buffer_paths.index(path)
        assert buffer == buffers[index]

    # and check that we can put it back together again
    put_buffers(state, buffer_paths, buffers)
    # we know that tuples get converted to list, so help the comparison by changing the
    # tuple to a list
    state_before["z"] = list(state_before["z"])
    assert state_before == state


def test_enables_widget_manager_in_colab(monkeypatch: pytest.MonkeyPatch):
    mock = MagicMock()
    monkeypatch.setitem(sys.modules, "google.colab.output", mock)
    get_repr_metadata()
    get_repr_metadata()
    mock.enable_custom_widget_manager.assert_called_once()


def test_get_metadata(monkeypatch: pytest.MonkeyPatch):
    meta = get_repr_metadata()
    assert meta == {}

    mock = MagicMock()
    mock._widgets._installed_url = None
    monkeypatch.setitem(sys.modules, "google.colab.output", mock)
    meta = get_repr_metadata()
    assert meta == {}

    mock = MagicMock()
    mock._widgets._installed_url = "foo"
    monkeypatch.setitem(sys.modules, "google.colab.output", mock)
    meta = get_repr_metadata()
    assert meta == {
        "application/vnd.jupyter.widget-view+json": {
            "colab": {"custom_widget_manager": {"url": "foo"}}
        }
    }


def test_try_file_contents_development(tmp_path: pathlib.Path):
    foo = tmp_path / "foo.txt"

    with pytest.raises(FileNotFoundError):
        try_file_contents(foo)

    foo.write_text("foo")
    assert try_file_contents(foo) is not None


def test_try_file_contents_development_enable_hmr(tmp_path: pathlib.Path):
    foo = tmp_path / "foo.txt"

    with pytest.raises(FileNotFoundError):
        try_file_contents(foo)

    foo.write_text("foo")
    with enable_hmr():
        file_contents = try_file_contents(foo)
    assert isinstance(file_contents, FileContents)
    assert file_contents._background_thread is not None
    file_contents.stop_thread()  # stop the background thread for CI


def test_try_file_contents_production(tmp_path: pathlib.Path):
    site_packages = tmp_path / "site-packages"
    site_packages.mkdir()
    bar = site_packages / "bar.txt"
    bar.write_text("bar")

    file_contents = try_file_contents(bar)
    assert isinstance(file_contents, FileContents)
    assert file_contents._background_thread is None


def test_try_file_contents_google_colab(tmp_path: pathlib.Path):
    site_packages = tmp_path / "dist-packages"
    site_packages.mkdir()
    bar = site_packages / "bar.txt"
    bar.write_text("bar")

    file_contents = try_file_contents(bar)
    assert isinstance(file_contents, FileContents)
    assert file_contents._background_thread is None


def test_try_file_contents_warns(
    monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path
):
    monkeypatch.setitem(sys.modules, "watchfiles", None)
    foo = tmp_path / "foo.txt"
    foo.write_text("foo")

    with enable_hmr():
        with pytest.warns(UserWarning, match="anywidget:"):
            file_contents = try_file_contents(foo)

        assert isinstance(file_contents, FileContents)
        assert file_contents._background_thread is None



from __future__ import annotations

from typing import Any

import ipywidgets
import traitlets.traitlets as t

from ._file_contents import FileContents
from ._util import (
    _ANYWIDGET_ID_KEY,
    _CSS_KEY,
    _DEFAULT_ESM,
    _ESM_KEY,
    enable_custom_widget_manager_once,
    get_repr_metadata,
    in_colab,
    try_file_contents,
)
from ._version import __version__
from .experimental import _register_anywidget_commands


class AnyWidget(ipywidgets.DOMWidget):  # type: ignore [misc]
    """Main AnyWidget base class."""

    _model_name = t.Unicode("AnyModel").tag(sync=True)
    _model_module = t.Unicode("anywidget").tag(sync=True)
    _model_module_version = t.Unicode(__version__).tag(sync=True)

    _view_name = t.Unicode("AnyView").tag(sync=True)
    _view_module = t.Unicode("anywidget").tag(sync=True)
    _view_module_version = t.Unicode(__version__).tag(sync=True)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if in_colab():
            enable_custom_widget_manager_once()

        anywidget_traits = {}
        for key in (_ESM_KEY, _CSS_KEY):
            if hasattr(self, key) and not self.has_trait(key):
                value = getattr(self, key)
                anywidget_traits[key] = t.Unicode(str(value)).tag(sync=True)
                if isinstance(value, FileContents):
                    value.changed.connect(
                        lambda new_contents, key=key: setattr(self, key, new_contents)
                    )

        # show default _esm if not defined
        if not hasattr(self, _ESM_KEY):
            anywidget_traits[_ESM_KEY] = t.Unicode(_DEFAULT_ESM).tag(sync=True)

        # TODO: a better way to uniquely identify this subclasses?
        # We use the fully-qualified name to get an id which we
        # can use to update CSS if necessary.
        anywidget_traits[_ANYWIDGET_ID_KEY] = t.Unicode(
            f"{self.__class__.__module__}.{self.__class__.__name__}"
        ).tag(sync=True)

        self.add_traits(**anywidget_traits)
        super().__init__(*args, **kwargs)
        _register_anywidget_commands(self)

    def __init_subclass__(cls, **kwargs: dict) -> None:
        """Coerces _esm and _css to FileContents if they are files."""
        super().__init_subclass__(**kwargs)
        for key in (_ESM_KEY, _CSS_KEY) & cls.__dict__.keys():
            # TODO: Upgrate to := when we drop Python 3.7
            file_contents = try_file_contents(getattr(cls, key))
            if file_contents:
                setattr(cls, key, file_contents)

    if hasattr(ipywidgets.DOMWidget, "_repr_mimebundle_"):
        # ipywidgets v8
        def _repr_mimebundle_(self, **kwargs: dict) -> tuple[dict, dict] | None:
            mimebundle = super()._repr_mimebundle_(**kwargs)
            if mimebundle is None:
                return None
            return mimebundle, get_repr_metadata()



import dataclasses

import psygnal
from anywidget._descriptor import ReprMimeBundle
from anywidget.experimental import MimeBundleDescriptor, dataclass, widget


def test_decorator():
    esm = "export default { render({ model , el }) {} }"
    css = ".foo { color: red;}"

    @widget(esm=esm, css=css)
    @psygnal.evented
    @dataclasses.dataclass
    class Foo:
        bar: str = "baz"

    foo = Foo()

    assert isinstance(Foo._repr_mimebundle_, MimeBundleDescriptor)  # type: ignore
    assert isinstance(foo._repr_mimebundle_, ReprMimeBundle)  # type: ignore


def test_dataclass():
    esm = "export function render({ model , el }) {}"
    css = ".foo { color: red;}"

    @dataclass(esm=esm, css=css)
    class Foo:
        bar: str = "baz"

    foo = Foo()

    assert isinstance(Foo._repr_mimebundle_, MimeBundleDescriptor)  # type: ignore
    assert isinstance(foo._repr_mimebundle_, ReprMimeBundle)  # type: ignore



import pathlib
import time
import weakref
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar, Set, Union
from unittest.mock import MagicMock, patch

import anywidget._descriptor
import pytest
import watchfiles
from anywidget._descriptor import (
    _COMMS,
    _WIDGET_MIME_TYPE,
    MimeBundleDescriptor,
    ReprMimeBundle,
)
from anywidget._file_contents import FileContents
from watchfiles import Change

if TYPE_CHECKING:
    from anywidget._protocols import AnywidgetProtocol
    from ipykernel.comm import Comm


class MockComm(MagicMock):
    # The only thing we need to do is to be able to relay messages back to
    # msg_callback, which is set by the descriptor using the on_msg method.

    msg_callback = None

    def on_msg(self, cb):
        self.msg_callback = cb

    def handle_msg(self, msg):
        if self.msg_callback is not None:
            self.msg_callback(msg)


@pytest.fixture
def mock_comm():
    """Mock a comm object."""
    comm = MockComm()
    assert not _COMMS
    with patch.object(anywidget._descriptor, "open_comm", return_value=comm):
        yield comm
    assert not _COMMS


def _send_value(comm: "Comm", value: int) -> int:
    # test that the object responds to incoming messages
    comm.handle_msg(
        {"content": {"data": {"method": "update", "state": {"value": value}}}}
    )
    return value


def _assert_sends_update(wdg: "AnywidgetProtocol", comm: MagicMock, expect: int):
    # test that the comm sends update messages
    wdg._repr_mimebundle_.send_state({"value"})
    comm.send.assert_called_with(
        data={"method": "update", "state": {"value": expect}, "buffer_paths": []},
        buffers=[],
    )


def test_descriptor(mock_comm: MagicMock) -> None:
    """Test that the descriptor decorator makes a comm, and gets/sets state."""

    VAL = 1
    REPR = "FOO"

    class Foo:
        _repr_mimebundle_ = MimeBundleDescriptor(autodetect_observer=False)
        value: int = VAL

        def _get_anywidget_state(self, include: Union[Set[str], None]):
            return {"value": self.value}

        def __repr__(self) -> str:
            return REPR

    foo = Foo()
    mock_comm.send.assert_not_called()  # we haven't yet created a comm object

    repr_method = foo._repr_mimebundle_  # the comm is created here
    mock_comm.send.assert_called_once()
    assert isinstance(repr_method, ReprMimeBundle)
    bundle = repr_method()
    assert _WIDGET_MIME_TYPE in bundle[0]  # we can call it as usual
    assert len(bundle[1]) == 0

    # test that the comm sends update messages
    _assert_sends_update(foo, mock_comm, VAL)

    # test that the object responds to incoming messages
    assert _send_value(mock_comm, 3) == foo.value

    mock_comm.send.reset_mock()
    mock_comm.handle_msg({"content": {"data": {"method": "request_state"}}})
    mock_comm.send.assert_called()

    # uses the base class repr for plain text
    assert foo._repr_mimebundle_()[0]["text/plain"] == REPR


def test_state_setter(mock_comm: MagicMock):
    """Test that `_set_anywidget_state` is used when present."""
    mock = MagicMock()

    class Foo:
        _repr_mimebundle_ = MimeBundleDescriptor(autodetect_observer=False)

        def _get_anywidget_state(self, include: Union[Set[str], None]):
            return {}

        def _set_anywidget_state(self, state):
            mock(state)

    foo = Foo()
    foo._repr_mimebundle_
    state = {"value": 7}
    mock_comm.handle_msg({"content": {"data": {"method": "update", "state": state}}})
    mock.assert_called_once_with(state)


def test_state_setter_binary(mock_comm: MagicMock):
    """Test that `_set_anywidget_state` is used when present."""
    mock = MagicMock()

    class Foo:
        _repr_mimebundle_ = MimeBundleDescriptor(autodetect_observer=False)

        def _get_anywidget_state(self, include: Union[Set[str], None]):
            return {}

        def _set_anywidget_state(self, state):
            mock(state)

    foo = Foo()
    foo._repr_mimebundle_
    mock_comm.handle_msg(
        {
            "content": {
                "data": {"method": "update", "state": {}, "buffer_paths": [["value"]]}
            },
            "buffers": [b"hello"],
        }
    )
    mock.assert_called_once_with({"value": b"hello"})


def test_comm_cleanup():
    """Test that the comm is cleaned up when the object is deleted."""
    assert not _COMMS

    class Foo:
        _repr_mimebundle_ = MimeBundleDescriptor(autodetect_observer=False)

        def _get_anywidget_state(self, include: Union[Set[str], None]):
            return {}

    foo = Foo()
    foo_ref = weakref.ref(foo)
    id_foo = id(foo)
    assert id_foo not in _COMMS
    repr_obj = foo._repr_mimebundle_
    assert id_foo in _COMMS
    del foo  # this should trigger the cleanup

    assert not _COMMS  # the comm should be gone
    assert foo_ref() is None  # the ref should be dead

    # setting this just so that we can test the following exception
    repr_obj._autodetect_observer = True
    with pytest.raises(RuntimeError):
        repr_obj.sync_object_with_view()


def test_detect_observer():
    class Foo:
        _repr_mimebundle_ = MimeBundleDescriptor()

        def _get_anywidget_state(self, include: Union[Set[str], None]):
            return {}

    with pytest.warns(UserWarning, match="Could not find a notifier"):
        Foo()._repr_mimebundle_


def test_descriptor_on_slots() -> None:
    """Make sure that strict classes don't break the descriptor altogether."""

    class Foo:
        __slots__ = ()

        _repr_mimebundle_ = MimeBundleDescriptor(autodetect_observer=False)
        value: int = 1

        def _get_anywidget_state(self, include: Union[Set[str], None]):
            return {"value": self.value}

    with pytest.warns(UserWarning, match=".*is not weakrefable"):
        Foo()._repr_mimebundle_

    # this test has the potential to leave a comm in _COMMS, so we clear it
    # to avoid polluting other tests
    _COMMS.clear()


def test_descriptor_with_psygnal(mock_comm: MagicMock):
    """Test that the observer pattern is found on psygnal.evented dataclasses."""
    psygnal = pytest.importorskip("psygnal")

    @psygnal.evented
    @dataclass
    class Foo:
        value: int = 1
        _repr_mimebundle_ = MimeBundleDescriptor()

    foo = Foo()
    repr_obj = foo._repr_mimebundle_  # create the comm

    mock_comm.send.reset_mock()
    foo.value = 2
    assert foo.value == 2
    mock_comm.send.assert_called_once_with(
        data={"method": "update", "state": {"value": 2}, "buffer_paths": []},
        buffers=[],
    )

    with pytest.warns(UserWarning, match="Refusing to re-sync a synced object"):
        repr_obj.sync_object_with_view()

    assert repr_obj._disconnectors
    del foo
    assert not repr_obj._disconnectors


def test_descriptor_with_pydantic(mock_comm: MagicMock):
    if TYPE_CHECKING:
        import pydantic
    else:
        pydantic = pytest.importorskip("pydantic")

    VAL = 1

    class Foo(pydantic.BaseModel):
        __slots__ = ("__weakref__",)
        value: int = VAL

        _repr_mimebundle_: ClassVar = MimeBundleDescriptor(autodetect_observer=False)

    foo = Foo()
    foo._repr_mimebundle_  # create the comm

    # test that the comm sends update messages
    _assert_sends_update(foo, mock_comm, VAL)

    # test that the object responds to incoming messages
    assert _send_value(mock_comm, 3) == foo.value


def test_descriptor_with_msgspec(mock_comm: MagicMock):
    if TYPE_CHECKING:
        import msgspec
        import psygnal
    else:
        psygnal = pytest.importorskip("psygnal")
        msgspec = pytest.importorskip("msgspec")

    VAL = 1

    @psygnal.evented
    class Foo(msgspec.Struct, weakref=True):
        value: int = VAL
        _repr_mimebundle_: ClassVar = MimeBundleDescriptor(autodetect_observer=False)

    foo = Foo()
    foo._repr_mimebundle_  # create the comm

    # test that the comm sends update messages
    _assert_sends_update(foo, mock_comm, VAL)

    # test that the object responds to incoming messages
    assert _send_value(mock_comm, 3) == foo.value


def test_descriptor_with_traitlets(mock_comm: MagicMock):
    import traitlets

    class Foo(traitlets.HasTraits):
        value = traitlets.Int(0).tag(sync=True)
        _repr_mimebundle_ = MimeBundleDescriptor()

    foo = Foo()
    repr_obj = foo._repr_mimebundle_  # create the comm
    mock_comm.send.reset_mock()

    foo.value = 2
    assert foo.value == 2
    mock_comm.send.assert_called_once_with(
        data={"method": "update", "state": {"value": 2}, "buffer_paths": []},
        buffers=[],
    )

    assert repr_obj._disconnectors
    mock_comm.send.reset_mock()
    repr_obj.unsync_object_with_view()
    foo.value = 5
    mock_comm.assert_not_called()
    assert not repr_obj._disconnectors


def test_infer_file_contents(mock_comm: MagicMock, tmp_path: pathlib.Path) -> None:
    """Test that the file contents are inferred from the file path."""

    site_packages = tmp_path / "site-packages"
    site_packages.mkdir()

    esm = site_packages / "foo.js"
    esm.write_text(
        "export default { render({ model, el }) { el.innerText = 'Hello, world'; } }"
    )

    class Foo:
        _repr_mimebundle_ = MimeBundleDescriptor(_esm=esm, autodetect_observer=False)
        value: int = 1

        def _get_anywidget_state(self, include: Union[Set[str], None]):
            return {"value": self.value}

    file_contents = Foo._repr_mimebundle_._extra_state["_esm"]
    assert isinstance(file_contents, FileContents)
    assert file_contents._background_thread is None

    foo = Foo()
    assert foo._repr_mimebundle_._extra_state["_esm"] == esm.read_text()

    def mock_file_events():
        esm.write_text("blah")
        # write to file
        changes = set()
        changes.add((Change.modified, str(esm)))
        yield changes
        # delete the file
        changes = set()
        changes.add((Change.deleted, str(esm)))
        yield changes

    with patch.object(watchfiles, "watch") as mock_watch:
        mock_watch.return_value = mock_file_events()
        file_contents.watch_in_thread()

    while (
        file_contents._background_thread and file_contents._background_thread.is_alive()
    ):
        time.sleep(0.01)

    mock_comm.send.assert_called_with(
        data={"method": "update", "state": {"_esm": "blah"}, "buffer_paths": []},
        buffers=[],
    )


def test_explicit_file_contents(tmp_path: pathlib.Path) -> None:
    """Test that the file contents are inferred from the file path."""

    path = tmp_path / "bar.txt"
    path.write_text("Hello, world")

    bar = FileContents(path, start_thread=False)

    class Foo:
        _repr_mimebundle_ = MimeBundleDescriptor(bar=bar, autodetect_observer=False)
        value: int = 1

        def _get_anywidget_state(self, include: Union[Set[str], None]):
            return {"value": self.value}

    file_contents = Foo._repr_mimebundle_._extra_state["bar"]
    assert file_contents == bar
    assert file_contents._background_thread is None

    foo = Foo()
    assert foo._repr_mimebundle_._extra_state["bar"] == path.read_text()


def test_no_view():
    """Test that the descriptor works without a view."""

    esm = """
    export function initialize({ model }) {
        model.on("msg:custom", (msg) => console.log(msg));
    }
    """

    class Foo:
        _repr_mimebundle_ = MimeBundleDescriptor(
            _esm=esm, no_view=True, autodetect_observer=False
        )

        def _get_anywidget_state(self, include: Union[Set[str], None]):
            return {}

    foo = Foo()
    assert foo._repr_mimebundle_() is None



import pathlib
import time
from collections import deque
from unittest.mock import MagicMock, Mock, patch

import pytest
import watchfiles
from anywidget._file_contents import FileContents
from watchfiles import Change


def test_file_contents_no_watch(tmp_path: pathlib.Path):
    """Test __str__ reads file contents and does not start a thread"""
    CONTENTS = "hello, world"
    path = tmp_path / "foo.txt"
    path.write_text(CONTENTS)

    with patch.object(watchfiles, "watch") as mock:
        contents = FileContents(path, start_thread=False)
        assert str(contents) == CONTENTS
        assert contents._background_thread is None
    mock.assert_not_called()


def test_file_contents_deleted(tmp_path: pathlib.Path):
    """Test deleting a file emits a deleted signal and stops the watcher"""
    CONTENTS = "hello, world"
    path = tmp_path / "foo.txt"
    path.write_text(CONTENTS)

    contents = FileContents(path, start_thread=False)
    mock = Mock()
    contents.deleted.connect(mock)

    def mock_file_events():
        changes = set()
        changes.add((Change.deleted, str(path)))
        yield changes
        changes = set()
        changes.add((Change.modified, str(path)))
        yield changes

    with patch.object(watchfiles, "watch") as mock_watch:
        mock_watch.return_value = mock_file_events()
        total = sum(1 for _ in contents.watch())

    assert total == 0
    mock_watch.assert_called_with(path, stop_event=contents._stop_event)
    assert mock.called


def test_file_contents_changed(tmp_path: pathlib.Path):
    """Test file changes emit changed signals and update the string contents"""
    CONTENTS = "hello, world"
    path = tmp_path / "foo.txt"
    path.write_text(CONTENTS)
    contents = FileContents(path, start_thread=False)

    mock = MagicMock()
    contents.changed.connect(mock)

    NEW_CONTENTS = "blah"

    def mock_file_events():
        path.write_text(NEW_CONTENTS)
        changes = set()
        changes.add((Change.modified, str(path)))
        yield changes

    with patch.object(watchfiles, "watch") as mock_watch:
        mock_watch.return_value = mock_file_events()
        deque(contents.watch(), maxlen=0)

    mock.assert_called_with(NEW_CONTENTS)
    assert str(contents) == NEW_CONTENTS


def test_file_contents_thread(tmp_path: pathlib.Path):
    """Test runs watcher in background thread by default"""
    path = tmp_path / "foo.txt"
    path.touch()

    contents = FileContents(path)

    # thread on by default
    assert contents._background_thread
    assert not contents._stop_event.is_set()
    assert contents._background_thread.is_alive()

    # reuse the same thread
    thread = contents._background_thread
    contents.watch_in_thread()
    assert contents._background_thread == thread

    # stops the thread
    contents.stop_thread()
    assert contents._stop_event.is_set()
    assert contents._background_thread is None

    # no-op
    contents.stop_thread()


def test_background_file_contents(tmp_path: pathlib.Path):
    """Test background thread watcher sends signals and updates contents"""
    CONTENTS = "hello, world"
    path = tmp_path / "foo.txt"
    path.write_text(CONTENTS)

    contents = FileContents(path, start_thread=False)
    mock_changed = MagicMock()
    contents.changed.connect(mock_changed)

    mock_deleted = MagicMock()
    contents.deleted.connect(mock_deleted)

    NEW_CONTENTS = "blah"

    def mock_file_events():
        # write to file
        path.write_text(NEW_CONTENTS)
        changes = set()
        changes.add((Change.modified, str(path)))
        yield changes
        # delete the file
        changes = set()
        changes.add((Change.deleted, str(path)))
        yield changes
        # "re-create the file"
        path.write_text(NEW_CONTENTS)
        changes = set()
        changes.add((Change.modified, str(path)))

    with patch.object(watchfiles, "watch") as mock_watch:
        mock_watch.return_value = mock_file_events()
        contents.watch_in_thread()

    while contents._background_thread and contents._background_thread.is_alive():
        time.sleep(0.01)

    mock_changed.assert_called_once_with(NEW_CONTENTS)
    assert str(contents) == NEW_CONTENTS


def test_missing_file_fails():
    """Test missing file fails to construct"""
    with pytest.raises(ValueError):
        FileContents("not_a_file.txt")

    with pytest.raises(ValueError):
        FileContents(pathlib.Path("not_a_file.txt"))



#!/usr/bin/env python

import json
import sys

import nbclient
import nbformat


def main():
    nb = nbformat.read(sys.argv[1], as_version=4)
    client = nbclient.NotebookClient(nb)
    client.execute()
    print(json.dumps(nb))


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import pathlib
import sys
import time
import typing
from unittest.mock import MagicMock, patch

import anywidget
import pytest
import traitlets.traitlets as t
import watchfiles
from anywidget._file_contents import FileContents
from anywidget._util import _DEFAULT_ESM, _WIDGET_MIME_TYPE
from traitlets import traitlets
from watchfiles import Change

here = pathlib.Path(__file__).parent


def enable_hmr():
    return patch.dict("os.environ", {"ANYWIDGET_HMR": "1"}, clear=True)


def test_version():
    with open(here / "../packages/anywidget/package.json") as f:
        pkg = json.load(f)

    assert anywidget.__version__ == pkg["version"]


def test_basic():
    ESM = """
    function render({ model, el }) {
        el.innerText = "Hello, world";
    }
    export default { render };
    """

    class Widget(anywidget.AnyWidget):
        _esm = t.Unicode(ESM).tag(sync=True)

    w = Widget()

    assert w.has_trait("_esm")
    assert w._esm == ESM


def test_default_esm():
    class Widget(anywidget.AnyWidget): ...

    w = Widget()

    assert w.has_trait("_esm")
    assert w._esm == _DEFAULT_ESM


def test_creates_fully_qualified_identifier():
    ESM = """
    function render({ model, el }) {
        el.innerText = "Hello, world";
    }
    export default { render };
    """

    class Widget(anywidget.AnyWidget):
        _module = t.Unicode(ESM).tag(sync=True)

    w = Widget()

    assert w.has_trait("_anywidget_id")
    assert w._anywidget_id == "test_widget.Widget"


def test_infer_traitlets():
    CSS = """
    .foo { background-color: black; }
    """

    ESM = """
    function render({ model, el }) {
        el.innerText = "Hello, world";
    }
    export default { render };
    """

    class Widget(anywidget.AnyWidget):
        _esm = ESM
        _css = CSS

    w = Widget()

    assert w.has_trait("_esm")
    assert w.trait_metadata("_esm", "sync")

    assert w.has_trait("_css")
    assert w.trait_metadata("_css", "sync")


def test_infer_traitlets_partial():
    CSS = """
    .foo { background-color: black; }
    """

    ESM = """
    function render({ model, el }) {
        el.innerText = "Hello, world";
    }
    export default { render };
    """

    class Widget(anywidget.AnyWidget):
        _esm = t.Unicode(ESM).tag(foo="bar")
        _css = CSS

    w = Widget()

    assert w.has_trait("_esm")
    assert w._esm == ESM
    assert w.trait_metadata("_esm", "foo") == "bar"

    assert w.has_trait("_css")
    assert w._css == CSS
    assert w.trait_metadata("_css", "sync")


def test_patched_repr_ipywidget_v8(monkeypatch: pytest.MonkeyPatch):
    w = anywidget.AnyWidget()
    bundle = w._repr_mimebundle_()
    assert bundle[0] and _WIDGET_MIME_TYPE in bundle[0]
    assert bundle[1] == {}


def test_patched_repr_ipywidget_v8_colab(monkeypatch: pytest.MonkeyPatch):
    mock = MagicMock()
    mock._widgets._installed_url = "foo"
    monkeypatch.setitem(sys.modules, "google.colab.output", mock)

    w = anywidget.AnyWidget()
    bundle = w._repr_mimebundle_()
    assert bundle[0] and _WIDGET_MIME_TYPE in bundle[0]
    assert bundle[1] == {
        _WIDGET_MIME_TYPE: {
            "colab": {"custom_widget_manager": {"url": mock._widgets._installed_url}},
        }
    }


def test_infer_file_contents(tmp_path: pathlib.Path):
    esm = tmp_path / "foo.js"
    esm.write_text(
        "export default { render({ model, el }) { el.innerText = 'Hello, world'; } }"
    )

    site_packages = tmp_path / "site-packages"
    site_packages.mkdir()
    css = site_packages / "styles.css"
    css.write_text(".foo { background-color: black; }")

    with enable_hmr():

        class Widget(anywidget.AnyWidget):
            _esm = esm
            _css = str(css)

    assert isinstance(Widget._esm, FileContents)
    assert Widget._esm._background_thread is not None

    assert isinstance(Widget._css, FileContents)
    assert Widget._css._background_thread is None

    w = Widget()

    assert w.has_trait("_esm")
    assert w._esm == esm.read_text()

    assert w.has_trait("_css")
    assert w._css == css.read_text()

    def mock_file_events():
        css.write_text("blah")
        # write to file
        changes = set()
        changes.add((Change.modified, str(css)))
        yield changes
        # delete the file
        changes = set()
        changes.add((Change.deleted, str(css)))
        yield changes

    with patch.object(watchfiles, "watch") as mock_watch:
        mock_watch.return_value = mock_file_events()
        Widget._css.watch_in_thread()

    while Widget._css._background_thread and Widget._css._background_thread.is_alive():
        time.sleep(0.01)

    assert w._css == css.read_text()

    # need to teardown the thread for CI
    Widget._esm.stop_thread()


def test_missing_pathlib_path_raises(tmp_path: pathlib.Path):
    esm = tmp_path / "foo.js"

    with pytest.raises(FileNotFoundError):

        class Widget(anywidget.AnyWidget):
            _esm = esm


def test_missing_string_path_with_suffix_raises(tmp_path: pathlib.Path):
    str_path_with_suffix = str(tmp_path / "foo.js")

    with pytest.raises(FileNotFoundError):

        class Widget(anywidget.AnyWidget):
            _esm = str_path_with_suffix


def test_remote_contents():
    esm = "http://example.com/foo.js"
    css = "https://example.com/bar.css"

    class Widget(anywidget.AnyWidget):
        _esm = esm
        _css = css

    widget = Widget()
    assert widget._esm == esm
    assert widget._css == css


def test_missing_string_path_without_suffix_is_raw_string(tmp_path: pathlib.Path):
    str_path_without_suffix = str(tmp_path / "foo")

    class Widget(anywidget.AnyWidget):
        _esm = str_path_without_suffix

    assert Widget()._esm == str_path_without_suffix


def test_explicit_file_contents(tmp_path: pathlib.Path):
    path = tmp_path / "foo.js"
    path.write_text(
        "export default { render({ model, el }) { el.innerText = 'Hello, world'; } }"
    )
    esm = FileContents(path, start_thread=False)

    class Widget(anywidget.AnyWidget):
        _esm = esm

    assert Widget._esm == esm

    w = Widget()
    assert w._esm == path.read_text()


def test_dom_less_widget():
    class Widget(anywidget.AnyWidget):
        _view_name = traitlets.Any(None).tag(sync=True)
        _esm = """
        function render({ model, el }) {
            el.innerText = "Hello, world";
        }
        export default { render };
        """

    assert Widget()._repr_mimebundle_() is None


def test_command_not_registered_by_default():
    class Widget(anywidget.AnyWidget):
        _esm = "export default { render({ model, el }) { el.innerText = 'Hello, world'; } }"

    w = Widget()
    assert len(w._msg_callbacks.callbacks) == 0


def test_anywidget_commands_register_one_callback():
    import anywidget.experimental

    class Widget(anywidget.AnyWidget):
        _esm = """
        export default {
            async render({ model, el, experimental }) {
                let [msg] = await experimental.invoke("_echo", "hi");
                let [msg] = await experimental.invoke("_echo2", "hi");
            }
        }
        """

        @anywidget.experimental.command
        def _echo(
            self, msg: typing.Any, buffers: list[bytes]
        ) -> tuple[str, list[bytes]]:
            return msg, buffers

        @anywidget.experimental.command
        def _echo2(
            self, msg: typing.Any, buffers: list[bytes]
        ) -> tuple[str, list[bytes]]:
            return msg, buffers

    w = Widget()
    assert len(w._msg_callbacks.callbacks) == 1



try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    from importlib_metadata import PackageNotFoundError, version  # type: ignore

try:
    __version__ = version("anywidget")
except PackageNotFoundError:
    __version__ = "uninstalled"
